﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.Linq;

namespace FormulaTranslation
{
    public class ExpressionItem
    {
        public ExpressionItem(string strExpression)
        {
            _strExpression = strExpression.Trim();

            if (_strExpression.Length==0)
                throw new Exception("Invalid expression.");

            char[] cExp = _strExpression.ToCharArray();

            int iParenthesisDepth = 0; //括号深度(层数)
            ExpType prevSign = ExpType.First;
            int iHead = 0;
            bool bToSeparate = false; //表示存在加减或乘除要拆分
            int i;

            //先拆加减
            for (i = 0; i < cExp.Length; i++)
            {
                if (cExp[i] == '(')
                {
                    iParenthesisDepth++;
                }
                if (cExp[i] == ')')
                {
                     if (iParenthesisDepth > 0)
                     {
                        iParenthesisDepth--;
                     }
                     else
                     {
                         throw new Exception("Invalid ')' in : " + _strExpression);
                     }
                }
                if (iParenthesisDepth != 0)
                {
                    continue;
                }

                if (cExp[i]=='+' || cExp[i]=='-')
                {
                    _subExpressions.Add(new ExpressionItem(_strExpression.Substring(iHead, i - iHead)) { Type = prevSign });
                    bToSeparate = true;
                    prevSign = cExp[i]=='+'?ExpType.Plus:ExpType.Minus;
                    iHead = i + 1;
                }
            }

            if (iParenthesisDepth != 0)
            {
                throw new Exception("Missing ')' in : " + _strExpression);
            }

            //如果括号不配对，应该走不到这里，上一步会抛出异常

            if (bToSeparate)
            {
                _subExpressions.Add(new ExpressionItem(_strExpression.Substring(iHead, cExp.Length - iHead)) { Type = prevSign });
                return;
            }

            //再拆乘和除
            prevSign = ExpType.First; //前一符号
            iHead = 0;
            for(i=0; i<cExp.Length; i++)
            {

                if (cExp[i] == '(')
                {
                    iParenthesisDepth++;
                }
                if (cExp[i] == ')')
                {
                    iParenthesisDepth--;
                }
                if (iParenthesisDepth != 0)
                {
                    continue;
                }

                if (cExp[i]=='*' || cExp[i]=='/')
                {
                    _subExpressions.Add(new ExpressionItem(_strExpression.Substring(iHead, i - iHead)){Type = prevSign});
                    bToSeparate = true;
                    prevSign = cExp[i] == '*' ? ExpType.Multiply : ExpType.Divide;
                    iHead = i + 1;
                }
            }

            if(bToSeparate)
            {
                _subExpressions.Add(new ExpressionItem(_strExpression.Substring(iHead, cExp.Length - iHead)){Type = prevSign});
                return;
            }

            //再拆括号和函数
            int iLeftBracketIndex = _strExpression.IndexOf('(');
            if (_strExpression.LastIndexOf(')') == _strExpression.Length - 1)
            {
                if (iLeftBracketIndex > 0) //是一个函数
                {
                    string strFuncName = _strExpression.Substring(0, iLeftBracketIndex).Trim();
                    if (!IsValidFunctionName(strFuncName))
                        throw new Exception("Function name is illeagal: " + strFuncName);
                    _function = GetFunDelegate(strFuncName);

                    //拆分函数参数
                    string strArguments = _strExpression.Substring(iLeftBracketIndex + 1, _strExpression.Length - iLeftBracketIndex - 2);
                    iHead = 0;
                    for (i = 0; i < strArguments.Length; i++)
                    {
                        if (strArguments[i] == '(')
                            iParenthesisDepth++;
                        if (strArguments[i] == ')')
                            iParenthesisDepth--;
                        if (iParenthesisDepth != 0)
                            continue;
                        if (strArguments[i] == ',')
                        {
                            _subExpressions.Add(new ExpressionItem(strArguments.Substring(iHead, i - iHead)) { Type = ExpType.Argument });
                            iHead = i + 1;
                        }
                    }
                    _subExpressions.Add(new ExpressionItem(strArguments.Substring(iHead, strArguments.Length - iHead)) { Type = ExpType.Argument });
                    return;
                }
                else if (iLeftBracketIndex == 0) //无函数
                {
                    _subExpressions.Add(new ExpressionItem(_strExpression.Substring(1, _strExpression.Length - 2)) {Type = ExpType.First});
                    return;
                }
            }
            
            //是否变量？
            if (cExp[0]=='#')
            {
                int oi;
                if(!int.TryParse(_strExpression.Substring(1), out oi))
                    throw new Exception("Invalid variable expression : " + _strExpression);
                _iVariableIndex = oi;
                return;
            }

            //直接计算显式值
            double ov;
            if(!double.TryParse(_strExpression, out ov))
                throw new Exception("Invalid expression : " + _strExpression);
            _dDirectValue = ov;
        }
        public double Val(double[] args)
        {
            double? dRtn = null;

            if (_iVariableIndex!=null)
            {
                if (args==null || _iVariableIndex>=args.Length)
                {
                    throw new Exception("Invalid argument(s).");
                }
                return args[(int) _iVariableIndex];
            }

            if (_dDirectValue != null)
                return _dDirectValue.Value;

            foreach (ExpressionItem exp in _subExpressions)
            {
                double dItemVal = exp.Val(args);

                switch (exp.Type)
                {
                    case ExpType.Argument:
                        _listArgs.Add(dItemVal);
                        break;
                    case ExpType.First:
                        dRtn = dItemVal;
                        break;
                    case ExpType.Plus:
                        dRtn += dItemVal;
                        break;
                    case ExpType.Minus:
                        dRtn -= dItemVal;
                        break;
                    case ExpType.Multiply:
                        dRtn *= dItemVal;
                        break;
                    case ExpType.Divide:
                        dRtn /= dItemVal;
                        break;
                }
            }

            if (_function != null)
            {
                dRtn = _function(_listArgs.ToArray());
                _listArgs.Clear();
                return dRtn.Value;
            }

            if (dRtn == null)
                return 0.0; //表达式是函数参数，返回值无所谓

            return dRtn.Value;
        }
        private readonly List<double> _listArgs = new List<double>();
        private static bool IsValidFunctionName(string strFuncName)
        {
            if (strFuncName.Length == 0)
                return false;
            char[] cArr = strFuncName.ToCharArray();
            char cFist = cArr[0];
            if (!(cFist >= 'a' && cFist <= 'z' || cFist >= 'A' && cFist <= 'Z'))
                return false;
            return cArr.All(c => c > 'a' && c < 'z' || c > 'A' && c < 'Z' || c > '0' && c < '9' || c == '_');
        }
        private static FormulaFunc GetFunDelegate(string strFunName) //根据函数名称生成函数委派
        {
            Type funcs = typeof(Functions);
            MethodInfo mi = funcs.GetMethod(strFunName);
            if(mi==null)
                throw new Exception("Function does not exist: "+strFunName);
            return (FormulaFunc)Delegate.CreateDelegate(typeof(FormulaFunc), mi);
        }
        private enum ExpType
        {
            First,
            Plus,
            Minus,
            Multiply,
            Divide,
            Argument
        }
        private readonly string _strExpression;
        private double? _dDirectValue { get; set; } //表达式是直接值(标量)
        private int? _iVariableIndex { get; set; }  //表达式是传入参数(变量)
        private FormulaFunc _function { get; set; } //表达式是函数
        private ExpType Type { get; set; }          //表达式类型：首项、加、减、乘、除和参数
        private readonly List<ExpressionItem> _subExpressions = new List<ExpressionItem>(); //子表达式
    }

    class Program
    {
        static void Main()
        {
            Console.WriteLine("支持加、减、乘、除、括号和函数，使用“#n”来表示第n个参数，n从0开始");
            do
            {
                Console.WriteLine("\n--------------------------\n请输入表达式");
                string strExpression = Console.ReadLine();
                if (strExpression=="q")
                    break;
                ExpressionItem exp;
                try
                {
                    exp = new ExpressionItem(strExpression);
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message);
                    continue;
                }

            TRY_ARGUMENT_AGAIN:

                List<double> listArgs = new List<double>();
                if (strExpression.IndexOf('#') != -1)
                {
                    Console.WriteLine("请输入参数，参数之间用空格分隔");
                    string strArguments = Console.ReadLine();
                    string[] arrArguments = strArguments.Split(new[] {' '}, StringSplitOptions.RemoveEmptyEntries);
                    foreach (string s in arrArguments)
                    {
                        double d;
                        if (!double.TryParse(s, out d))
                        {
                            Console.WriteLine("参数不合法，请重试");
                            goto TRY_ARGUMENT_AGAIN;
                        }
                        listArgs.Add(d);
                    }
                }

                try
                {
                    Console.WriteLine("表达式计算结果: " + exp.Val(listArgs.ToArray()));
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message);
                }
            } while (true);
        }
    }
}
